/*
	Rest Query (shared lib)

	Copyright (c) 2014 - 2020 Cédric Ronvel

	The MIT License (MIT)

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all
	copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
	SOFTWARE.
*/

"use strict" ;



// Charmap for string validation

const charmap = {
	lowerCaseArray: [ 'a' , 'b' , 'c' , 'd' , 'e' , 'f' , 'g' , 'h' , 'i' , 'j' , 'k' , 'l' , 'm' , 'n' , 'o' , 'p' , 'q' , 'r' , 's' , 't' , 'u' , 'v' , 'w' , 'x' , 'y' , 'z' ] ,
	upperCaseArray: [ 'A' , 'B' , 'C' , 'D' , 'E' , 'F' , 'G' , 'H' , 'I' , 'J' , 'K' , 'L' , 'M' , 'N' , 'O' , 'P' , 'Q' , 'R' , 'S' , 'T' , 'U' , 'V' , 'W' , 'X' , 'Y' , 'Z' ] ,
	digitArray: [ '0' , '1' , '2' , '3' , '4' , '5' , '6' , '7' , '8' , '9' ] ,
	lowerCaseAndDigitArray: [ 'a' , 'b' , 'c' , 'd' , 'e' , 'f' , 'g' , 'h' , 'i' , 'j' , 'k' , 'l' , 'm' , 'n' , 'o' , 'p' , 'q' , 'r' , 's' , 't' , 'u' , 'v' , 'w' , 'x' , 'y' , 'z' , '0' , '1' , '2' , '3' , '4' , '5' , '6' , '7' , '8' , '9' ] ,
	collectionRegExp: /^[A-Z][a-zA-Z0-9]*$/ ,
	methodRegExp: /^[A-Z][A-Z0-9-]*$/ ,
	propertyRegExp: /^(\.[a-zA-Z0-9_-]+)+$/ ,
	linkPropertyRegExp: /^~[a-zA-Z0-9_-]+(\.[a-zA-Z0-9_-]+)*$/ ,
	multiLinkPropertyRegExp: /^~~[a-zA-Z0-9_-]+$/ ,
	idRegExp: /^[0-9a-f]{24}$/ ,
	rangeRegExp: /^([0-9]+)(?:-([0-9]+))?$/ ,
	slugIdRegExp: /^[a-z0-9-]{1,72}$/ ,

	// Unicode slugs, allowing: lowercase/modifier/other letter (excluding uppercase and titlecase), numbers, marks (accent, needed for arabic).
	// Currency symbols must be excluded because $ is considered unsafe in URLs
	unicodeSlugIdRegExp: /^[\p{Ll}\p{Lm}\p{Lo}\p{N}\p{M}-]{1,72}$/u ,

	// Slugify map
	// From Django urlify.js
	// Arabic from https://github.com/sindresorhus/slugify/blob/master/replacements.js
	asciiMapCommon: {
		// latin
		'À': 'A' ,
		'Á': 'A' ,
		'Â': 'A' ,
		'Ã': 'A' ,
		'Ä': 'A' ,
		'Å': 'A' ,
		'Æ': 'AE' ,
		'Ç': 'C' ,
		'È': 'E' ,
		'É': 'E' ,
		'Ê': 'E' ,
		'Ë': 'E' ,
		'Ì': 'I' ,
		'Í': 'I' ,
		'Î': 'I' ,
		'Ï': 'I' ,
		'Ð': 'D' ,
		'Ñ': 'N' ,
		'Ò': 'O' ,
		'Ó': 'O' ,
		'Ô': 'O' ,
		'Õ': 'O' ,
		'Ö': 'O' ,
		'Ő': 'O' ,
		'Ø': 'O' ,
		'Ù': 'U' ,
		'Ú': 'U' ,
		'Û': 'U' ,
		'Ü': 'U' ,
		'Ű': 'U' ,
		'Ý': 'Y' ,
		'Þ': 'TH' ,
		'ß': 'ss' ,
		'à': 'a' ,
		'á': 'a' ,
		'â': 'a' ,
		'ã': 'a' ,
		'ä': 'a' ,
		'å': 'a' ,
		'æ': 'ae' ,
		'ç': 'c' ,
		'è': 'e' ,
		'é': 'e' ,
		'ê': 'e' ,
		'ë': 'e' ,
		'ì': 'i' ,
		'í': 'i' ,
		'î': 'i' ,
		'ï': 'i' ,
		'ð': 'd' ,
		'ñ': 'n' ,
		'ò': 'o' ,
		'ó': 'o' ,
		'ô': 'o' ,
		'õ': 'o' ,
		'ö': 'o' ,
		'ő': 'o' ,
		'ø': 'o' ,
		'ù': 'u' ,
		'ú': 'u' ,
		'û': 'u' ,
		'ü': 'u' ,
		'ű': 'u' ,
		'ý': 'y' ,
		'þ': 'th' ,
		'ÿ': 'y' ,
		'ẞ': 'ss' ,
		'œ': 'oe' ,
		'Œ': 'OE' ,	// <-- moved from symbols to common
		// common
		'“': '"' ,
		'”': '"' ,
		'‘': "'" ,
		'’': "'" ,
		'…': '...'
	} ,
	// From Django urlify.js
	asciiMapWorldAlpha: {
		// greek
		'α': 'a' ,
		'β': 'b' ,
		'γ': 'g' ,
		'δ': 'd' ,
		'ε': 'e' ,
		'ζ': 'z' ,
		'η': 'h' ,
		'θ': '8' ,
		'ι': 'i' ,
		'κ': 'k' ,
		'λ': 'l' ,
		'μ': 'm' ,
		'ν': 'n' ,
		'ξ': '3' ,
		'ο': 'o' ,
		'π': 'p' ,
		'ρ': 'r' ,
		'σ': 's' ,
		'τ': 't' ,
		'υ': 'y' ,
		'φ': 'f' ,
		'χ': 'x' ,
		'ψ': 'ps' ,
		'ω': 'w' ,
		'ά': 'a' ,
		'έ': 'e' ,
		'ί': 'i' ,
		'ό': 'o' ,
		'ύ': 'y' ,
		'ή': 'h' ,
		'ώ': 'w' ,
		'ς': 's' ,
		'ϊ': 'i' ,
		'ΰ': 'y' ,
		'ϋ': 'y' ,
		'ΐ': 'i' ,
		'Α': 'A' ,
		'Β': 'B' ,
		'Γ': 'G' ,
		'Δ': 'D' ,
		'Ε': 'E' ,
		'Ζ': 'Z' ,
		'Η': 'H' ,
		'Θ': '8' ,
		'Ι': 'I' ,
		'Κ': 'K' ,
		'Λ': 'L' ,
		'Μ': 'M' ,
		'Ν': 'N' ,
		'Ξ': '3' ,
		'Ο': 'O' ,
		'Π': 'P' ,
		'Ρ': 'R' ,
		'Σ': 'S' ,
		'Τ': 'T' ,
		'Υ': 'Y' ,
		'Φ': 'F' ,
		'Χ': 'X' ,
		'Ψ': 'PS' ,
		'Ω': 'W' ,
		'Ά': 'A' ,
		'Έ': 'E' ,
		'Ί': 'I' ,
		'Ό': 'O' ,
		'Ύ': 'Y' ,
		'Ή': 'H' ,
		'Ώ': 'W' ,
		'Ϊ': 'I' ,
		'Ϋ': 'Y' ,
		// turkish
		'ş': 's' ,
		'Ş': 'S' ,
		'ı': 'i' ,
		'İ': 'I' ,
		'ğ': 'g' ,
		'Ğ': 'G' ,
		//  russian
		'а': 'a' ,
		'б': 'b' ,
		'в': 'v' ,
		'г': 'g' ,
		'д': 'd' ,
		'е': 'e' ,
		'ё': 'yo' ,
		'ж': 'zh' ,
		'з': 'z' ,
		'и': 'i' ,
		'й': 'j' ,
		'к': 'k' ,
		'л': 'l' ,
		'м': 'm' ,
		'н': 'n' ,
		'о': 'o' ,
		'п': 'p' ,
		'р': 'r' ,
		'с': 's' ,
		'т': 't' ,
		'у': 'u' ,
		'ф': 'f' ,
		'х': 'h' ,
		'ц': 'c' ,
		'ч': 'ch' ,
		'ш': 'sh' ,
		'щ': 'sh' ,
		'ъ': 'u' ,
		'ы': 'y' ,
		'ь': '' ,
		'э': 'e' ,
		'ю': 'yu' ,
		'я': 'ya' ,
		'А': 'A' ,
		'Б': 'B' ,
		'В': 'V' ,
		'Г': 'G' ,
		'Д': 'D' ,
		'Е': 'E' ,
		'Ё': 'Yo' ,
		'Ж': 'Zh' ,
		'З': 'Z' ,
		'И': 'I' ,
		'Й': 'J' ,
		'К': 'K' ,
		'Л': 'L' ,
		'М': 'M' ,
		'Н': 'N' ,
		'О': 'O' ,
		'П': 'P' ,
		'Р': 'R' ,
		'С': 'S' ,
		'Т': 'T' ,
		'У': 'U' ,
		'Ф': 'F' ,
		'Х': 'H' ,
		'Ц': 'C' ,
		'Ч': 'Ch' ,
		'Ш': 'Sh' ,
		'Щ': 'Sh' ,
		'Ъ': 'U' ,
		'Ы': 'Y' ,
		'Ь': '' ,
		'Э': 'E' ,
		'Ю': 'Yu' ,
		'Я': 'Ya' ,
		// ukranian
		'Є': 'Ye' ,
		'І': 'I' ,
		'Ї': 'Yi' ,
		'Ґ': 'G' ,
		'є': 'ye' ,
		'і': 'i' ,
		'ї': 'yi' ,
		'ґ': 'g' ,
		// czech
		'č': 'c' ,
		'ď': 'd' ,
		'ě': 'e' ,
		'ň': 'n' ,
		'ř': 'r' ,
		'š': 's' ,
		'ť': 't' ,
		'ů': 'u' ,
		'ž': 'z' ,
		'Č': 'C' ,
		'Ď': 'D' ,
		'Ě': 'E' ,
		'Ň': 'N' ,
		'Ř': 'R' ,
		'Š': 'S' ,
		'Ť': 'T' ,
		'Ů': 'U' ,
		'Ž': 'Z' ,
		// polish
		'ą': 'a' ,
		'ć': 'c' ,
		'ę': 'e' ,
		'ł': 'l' ,
		'ń': 'n' ,
		'ś': 's' ,
		'ź': 'z' ,
		'ż': 'z' ,
		'Ą': 'A' ,
		'Ć': 'C' ,
		'Ę': 'e' ,
		'Ł': 'L' ,
		'Ń': 'N' ,
		'Ś': 'S' ,
		'Ź': 'Z' ,
		'Ż': 'Z' ,
		// latvian
		'ā': 'a' ,
		'ē': 'e' ,
		'ģ': 'g' ,
		'ī': 'i' ,
		'ķ': 'k' ,
		'ļ': 'l' ,
		'ņ': 'n' ,
		'ū': 'u' ,
		'Ā': 'A' ,
		'Ē': 'E' ,
		'Ģ': 'G' ,
		'Ī': 'i' ,
		'Ķ': 'k' ,
		'Ļ': 'L' ,
		'Ņ': 'N' ,
		'Ū': 'u' ,

		// from https://github.com/sindresorhus/slugify/blob/master/replacements.js
		// arabic
		'ء': 'e' ,
		'آ': 'a' ,
		'أ': 'a' ,
		'ؤ': 'w' ,
		'إ': 'i' ,
		'ئ': 'y' ,
		'ا': 'a' ,
		'ب': 'b' ,
		'ة': 't' ,
		'ت': 't' ,
		'ث': 'th' ,
		'ج': 'j' ,
		'ح': 'h' ,
		'خ': 'kh' ,
		'د': 'd' ,
		'ذ': 'dh' ,
		'ر': 'r' ,
		'ز': 'z' ,
		'س': 's' ,
		'ش': 'sh' ,
		'ص': 's' ,
		'ض': 'd' ,
		'ط': 't' ,
		'ظ': 'z' ,
		'ع': 'e' ,
		'غ': 'gh' ,
		'ـ': '_' ,
		'ف': 'f' ,
		'ق': 'q' ,
		'ك': 'k' ,
		'ل': 'l' ,
		'م': 'm' ,
		'ن': 'n' ,
		'ه': 'h' ,
		'و': 'w' ,
		'ى': 'a' ,
		'ي': 'y' ,
		'َ‎': 'a' ,
		'ُ': 'u' ,
		'ِ‎': 'i' ,
		'٠': '0' ,
		'١': '1' ,
		'٢': '2' ,
		'٣': '3' ,
		'٤': '4' ,
		'٥': '5' ,
		'٦': '6' ,
		'٧': '7' ,
		'٨': '8' ,
		'٩': '9' ,

		// Persian / Farsi
		'چ': 'ch' ,
		'ک': 'k' ,
		'گ': 'g' ,
		'پ': 'p' ,
		'ژ': 'zh' ,
		'ی': 'y' ,
		'۰': '0' ,
		'۱': '1' ,
		'۲': '2' ,
		'۳': '3' ,
		'۴': '4' ,
		'۵': '5' ,
		'۶': '6' ,
		'۷': '7' ,
		'۸': '8' ,
		'۹': '9' ,

		// Pashto
		'ټ': 'p' ,
		'ځ': 'z' ,
		'څ': 'c' ,
		'ډ': 'd' ,
		'ﺫ': 'd' ,
		'ﺭ': 'r' ,
		'ړ': 'r' ,
		'ﺯ': 'z' ,
		'ږ': 'g' ,
		'ښ': 'x' ,
		'ګ': 'g' ,
		'ڼ': 'n' ,
		'ۀ': 'e' ,
		'ې': 'e' ,
		'ۍ': 'ai' ,

		// Urdu
		'ٹ': 't' ,
		'ڈ': 'd' ,
		'ڑ': 'r' ,
		'ں': 'n' ,
		'ہ': 'h' ,
		'ھ': 'h' ,
		'ے': 'e'
	} ,
	asciiMapSymbolsEn: {
		// currency
		'€': ' euro' ,
		'₢': ' cruzeiro' ,
		'₣': ' french franc' ,
		'£': ' pound' ,
		'₤': ' lira' ,
		'₥': ' mill' ,
		'₦': ' naira' ,
		'₧': ' peseta' ,
		'₨': ' rupee' ,
		'₩': ' won' ,
		'₪': ' new shequel' ,
		'₫': ' dong' ,
		'₭': ' kip' ,
		'₮': ' tugrik' ,
		'₯': ' drachma' ,
		'₰': ' penny' ,
		'₱': ' peso' ,
		'₲': ' guarani' ,
		'₳': ' austral' ,
		'₴': ' hryvnia' ,
		'₵': ' cedi' ,
		'¢': ' cent' ,
		'¥': ' yen' ,
		'元': ' yuan' ,
		'円': ' yen' ,
		'﷼': ' rial' ,
		'₠': ' ecu' ,
		'¤': ' currency' ,
		'฿': ' baht' ,
		"\\$": ' dollar' ,
		// symbols
		'©': ' (c)' ,
		'∑': ' sum' ,
		'®': ' (r)' ,
		'†': ' +' ,
		'∂': ' d' ,
		'ƒ': ' f' ,
		'™': ' tm' ,
		'℠': ' sm' ,
		'˚': ' o' ,
		'º': ' o' ,
		'ª': ' a' ,
		'•': ' *' ,
		'∆': ' delta' ,
		'∞': ' infinity' ,
		'♥': ' love' ,
		'&': ' and' ,
		'\\|': ' or' ,
		'<': ' less' ,
		'>': ' greater'
	} ,
	asciiMapSymbolsFr: {
		// currency
		'€': ' euro' ,
		'₢': ' cruzeiro' ,
		'₣': ' franc' ,
		'£': ' livre' ,
		'₤': ' lire' ,
		'₥': ' mill' ,
		'₦': ' naira' ,
		'₧': ' peseta' ,
		'₨': ' rupee' ,
		'₩': ' won' ,
		'₪': ' new shequel' ,
		'₫': ' dong' ,
		'₭': ' kip' ,
		'₮': ' tugrik' ,
		'₯': ' drachma' ,
		'₰': ' penny' ,
		'₱': ' peso' ,
		'₲': ' guarani' ,
		'₳': ' austral' ,
		'₴': ' hryvnia' ,
		'₵': ' cedi' ,
		'¢': ' cent' ,
		'¥': ' yen' ,
		'元': ' yuan' ,
		'円': ' yen' ,
		'﷼': ' rial' ,
		'₠': ' ecu' ,
		'¤': ' monnaie' ,
		'฿': ' baht' ,
		"\\$": ' dollar' ,
		// symbols
		'©': ' (c)' ,
		'∑': ' sum' ,
		'®': ' (r)' ,
		'†': ' +' ,
		'∂': ' d' ,
		'ƒ': ' f' ,
		'™': ' tm' ,
		'℠': ' sm' ,
		'˚': ' o' ,
		'º': ' o' ,
		'ª': ' a' ,
		'•': ' *' ,
		'∆': ' delta' ,
		'∞': ' infini' ,
		'♥': ' aime' ,
		'&': ' et' ,
		'\\|': ' ou' ,
		'<': ' moins' ,
		'>': ' plus'
	}
} ;

module.exports = charmap ;

